def digits(num):
    return list(map(int, str(num)))


def sumprod(a, b):
    r = 1
    for x, y in zip(a, b):
        r += x * y
    return r


def eulers_totient(num: int, *, pf: Dict[int, int] = None) -> int:
    phi = num
    for f in pf or prime_factors(num):
        phi *= (1 - 1/f)
    return int(round(phi))


def congruent(a, b, modulo) -> bool:
    return (a % modulo) == (b % modulo)


def chinese_remainder_theorem(remainders: list, moduli: list):
    """Solve a system of congruences, such as:
    x = 3   (mod 5)
    x = 1   (mod 7)
    x = 6   (mod 8)
    => find all values for x that satisfy this system
    """
    N = prod(moduli)
    nums = N // moduli
    inverses = [modular_inverse(num, modulo)
                for num, modulo in zip(nums, moduli)]
    x = sum(r*n*i for r, n, i in zip(remainders, nums, inverses)) % N
    return x, N



import numpy as np
from math import isqrt


def primes_below(n) -> np.ndarray:
    if n < 3:
        return np.array([])
    if n == 3:
        return np.array([3])
    sieve = np.ones(n//3 + (n % 6 == 2), dtype=np.bool_)
    sieve[0] = False
    for i in range(isqrt(n)//3+1):
        if sieve[i]:
            k = 3*i+1 | 1
            sieve[((k*k)//3)::2*k] = False
            sieve[(k*k+4*k-2*k*(i & 1))//3::2*k] = False
    return np.r_[2, 3, ((3*np.nonzero(sieve)[0]+1) | 1)]


def primes_between():
    raise NotImplementedError()


if __name__ == "__main__":

    q = primes_below(8)
    print(q)



def prime_factors(n: int) -> dict:
    """Inefficient use sieve instead."""
    if n < 2:
        return dict()

    prime_factors = dict()
    while n % 2 == 0:
        prime_factors[2] = prime_factors.get(2, 0) + 1
        n //= 2

    for factor in range(3, n + 1, 2):
        while n % factor == 0:
            prime_factors[factor] = prime_factors.get(factor, 0) + 1
            n //= factor
    return prime_factors



from ntheory.primes import prime_factors


def test_prime_factors():
    assert prime_factors(-10) == {}
    assert prime_factors(1) == {}
    assert prime_factors(2) == {2: 1}
    assert prime_factors(8) == {2: 3}
    assert prime_factors(8) == {2: 3}



from ntheory.modulo import modular_sum, modular_prod, modular_pow, gcd_extended, modular_inverse


def test_modular_sum():
    assert modular_sum(14, 17, mod=5) == 1


def test_modular_prod():
    assert modular_prod(4, 7, mod=6) == 4


def test_modular_pow():
    assert modular_pow(5, 117, mod=19) == 1


def test_gcd_extended():
    assert gcd_extended(888, 54) == (6, -2, 33)
    assert gcd_extended(1180, 482) == (2, -29, 71)
    assert gcd_extended(482, 1180) == (2, 71, -29)


def test_modular_inverse():
    assert modular_inverse(5, modulo=56) == 45
    assert modular_inverse(56, modulo=5) == 1
    assert modular_inverse(40, modulo=7) == 3
    assert modular_inverse(35, modulo=8) == 3
    assert modular_inverse(197, modulo=3000) == 533



from ..binary import binary_digits


def modular_sum(*args, mod):
    """Compute (a+b)%modulo"""
    r = 0
    for x in args:
        r = (r + x) % mod
    return r


def modular_prod(*args, mod):
    """Compute (a+b)%modulo"""
    r = 1
    for x in args:
        r = (r*x) % mod
    return r


def modular_pow(a, b, *, mod):
    """ Modular exponentiation a^b
    5 ** 117 = 5 ** (1 + 4 + 16 + 32 + 64)
    """
    factors = []
    for bit in binary_digits(b)[::-1]:
        if bit:
            factors.append(a)
        a = modular_prod(a, a, mod=mod)

    return modular_prod(*factors, mod=mod)


def gcd_extended(a, b):
    """Implement the extended euclidian algorithm
    a*x + b*y = gcd(a, b)
    """
    if b > a:
        gcd, x, y = gcd_extended(b, a)
        return gcd, y, x
        # a, b = b, a

    chart = []
    while b > 0:
        chart.insert(0, (a, a//b))
        a, b = b, a % b
    gcd = a

    a, x, b, y = 0, 1, 0, 0
    for p, s in chart:
        a, x, b, y = p, y, a, x - y*s
    return gcd, x, y


def modular_inverse(num, modulo):
    g, x, y = gcd_extended(num, modulo)
    if g != 1:
        raise ValueError('No inverse exists!')
    return x % modulo



from setuptools import find_packages, setup

setup(
    name='ntheory',
    packages=find_packages(),
    version='0.0.8',
    description='My first Python library',
    author='Kasper Arfman',
    license='MIT',
    install_requires=[
        'numpy==1.24.3',
    ],
    setup_requires=['pytest-runner'],
    tests_require=['pytest==7.3.1'],
    test_suite='tests',
)
